unit Compiler;

{ Smart Expression Compiler for Turbo/Borland Pascal 7.0 }

interface
{$X+,B-,N-,E-}
uses
   Objects;

type
   TIdParams =
   object(TObject)
      Name         :PString;

      constructor Init(const AName     :String);
      destructor Done; virtual;
   end;

   TGetValueProc = Function(const Str  :String;
                            var Val    :String)  :Boolean;

   TTypeParams =
   object(TObject)
      Id           :Word;
      Size         :Word;
      GetValue     :TGetValueProc;

      constructor Init(AnId            :Word;
                       ASize           :Word;
                       AGetValue       :TGetValueProc);
   end;

   TVarParams =
   object(TIdParams)
      Addr         :Pointer;
      TypeV        :Integer;

      constructor Init(const AName     :String;
                       var AVar;
                       AType           :Integer);
   end;

   TOpParams =
   object(TIdParams)
      Binary       :Boolean;
      Types        :Array [0..1] of Integer;
      TypeR        :Integer;
      Precedence   :Integer;
      Code         :Pointer;

      constructor Init(const AName     :String;
                       AType1, AType2,
                       ATypeR          :Integer;
                       APrecedence     :Integer;
                       ACode           :Pointer);
   end;

   TConvParams =
   object(TObject)
      TypeI, TypeO :Integer;
      Code         :Pointer;

      constructor Init(ATypeI, ATypeO  :Integer;
                       ACode           :Pointer);
   end;

   TIdsCollection =
   object(TSortedCollection)
      function  Compare(Key1, Key2     :Pointer) :Integer; virtual;
      function  Search(Key             :Pointer;
                       var Index       :Integer) :Boolean; virtual;
   end;

   TOpsCollection =
   object(TIdsCollection)
      function  Compare(Key1, Key2     :Pointer) :Integer; virtual;

   private
      CmpMethod    :(cmName, cmKind, cmTypes);
   end;

   TConvsCollection =
   object(TSortedCollection)
      function  Compare(Key1, Key2     :Pointer) :Integer; virtual;
   end;

   TExpressionCompiler =
   object(TObject)
      Status       :Integer;
      Types        :PCollection;
      Vars         :^TIdsCollection;
      Ops          :^TOpsCollection;
      Convs        :^TConvsCollection;

      CaseSensitive:Boolean;
      { This field can only be changed immediately after calling Init }

      SafeMode,
      Optimization :Boolean;
      { These fields can be changed at any time }

      Code         :Pointer;
      CodeSize     :Word;
      { These fields are set by Compile (if Status=ecOk) }

      constructor Init;
      destructor Done; virtual;

      procedure FreeCodeMem;

      procedure AddType(AnId           :Word;
                        ASize          :Word;
                        AGetValue      :TGetValueProc);

      procedure AddOp(const AName      :String;
                      AType1, AType2,
                      ATypeR           :Word;
                      APrecedence      :Integer;
                      ACode            :Pointer);

      procedure AddConvert(TypeI, TypeO:Word;
                           ACode       :Pointer);

      procedure AddVar(const AName     :String;
                       var AVar;
                       AType           :Word);

      procedure Compile(Expr           :String;
                        TypeR          :Word);

      procedure DefineInteger;
      procedure DefineReal;
      procedure DefineIR;

      procedure DefineBoolean;
      procedure DefineBI;
      procedure DefineBR;

      procedure DefineAll;

   private
      function  FindType(Id            :Word)  :Integer;
   end;

const
   { Error codes }
   ecOk =          0;
   ecIntError =    -1;
   ecError =       1;
   ecTooComplex =  2;
   ecUnknownId =   3;
   ecDuplicate =   4;
   ecInvValue =    5;
   ecInvType =     6;
   ecInvOpUsage =  7;

   { Type codes }
   tiNone =        0;
   tiInteger =     1;
   tiReal =        8;
   tiBoolean =     16;

implementation

const
   diEndExpr =     #0;
   diVarIndex =    #1;
   diOpIndex =     #2;
   diValueASCII =  #4;
   diValueInt =    #5;
   diStack =       #8;

const
   GCaseSensitive  :Boolean= False;
   NoCode          :Byte =   $CB;

procedure SLoCase(var s                :String); assembler;
asm
   les  di,s
   xor  cx,cx
   mov  cl,es:[di]
   jcxz @@3
@@1:
   inc  di
   mov  al,es:[di]
   cmp  al,'A'
   jb   @@2
   cmp  al,'Z'
   ja   @@2
   add  al,'a'-'A'
   mov  es:[di],al
@@2:
   loop @@1
@@3:
end;

type
   PInteger =      ^Integer;
   PLongInt =      ^LongInt;

{ TIdParams }

constructor TIdParams.Init;
begin
   Inherited Init;
   Name:=NewStr(AName);
   if not GCaseSensitive then SLoCase(Name^);
end;

destructor TIdParams.Done;
begin
   DisposeStr(Name);
end;

{ TTypeParams }

constructor TTypeParams.Init;
begin
   Inherited Init;
   Id:=AnId; Size:=ASize; GetValue:=AGetValue;
end;

{ TVarParams }

constructor TVarParams.Init;
begin
   Inherited Init(AName);
   Addr:=@AVar; TypeV:=AType;
end;

{ TOpParams }

constructor TOpParams.Init;
begin
   Inherited Init(AName);
   Types[0]:=AType2; Types[1]:=AType1; TypeR:=ATypeR;
   Binary:=(AType1<>0); Precedence:=APrecedence;
   Code:=ACode;
end;

{ TConvParams }

constructor TConvParams.Init;
begin
   Inherited Init;
   TypeI:=ATypeI; TypeO:=ATypeO;
   Code:=ACode;
end;

{ TIdsCollection }

function  TIdsCollection.Compare;

function  GetString(P                  :PString) :String;
begin
   if P=nil then GetString:='' else GetString:=P^;
end;

begin
   if GetString(TIdParams(Key1^).Name) < GetString(TIdParams(Key2^).Name) then Compare:=-1 else
   if GetString(TIdParams(Key1^).Name) > GetString(TIdParams(Key2^).Name) then Compare:=+1 else
   Compare:=0;
end;

function  TIdsCollection.Search;
var
   SaveName        :String;
begin
   if not GCaseSensitive then
   begin
      SaveName:=TIdParams(Key^).Name^;
      SLoCase(TIdParams(Key^).Name^);
   end;
   Search:=Inherited Search(Key, Index);
   if not GCaseSensitive then TIdParams(Key^).Name^:=SaveName;
end;

{ TOpsCollection }

function  TOpsCollection.Compare;
var
   LastCmp         :Integer;
begin
   LastCmp:=Inherited Compare(Key1, Key2);
   if (LastCmp<>0) or (CmpMethod=cmName) then Compare:=LastCmp else
   if CmpMethod=cmKind then
   begin
      if TOpParams(Key1^).Binary < TOpParams(Key2^).Binary then Compare:=-1 else
      if TOpParams(Key1^).Binary > TOpParams(Key2^).Binary then Compare:=+1 else
      Compare:=0;
   end else
   begin
      CmpMethod:=cmKind;
      LastCmp:=Compare(Key1, Key2);
      CmpMethod:=cmTypes;
      if LastCmp<>0 then Compare:=LastCmp else
      begin
         if TOpParams(Key1^).Types[0] < TOpParams(Key2^).Types[0] then Compare:=-1 else
         if TOpParams(Key1^).Types[0] > TOpParams(Key2^).Types[0] then Compare:=+1 else
         if TOpParams(Key1^).Types[1] < TOpParams(Key2^).Types[1] then Compare:=-1 else
         if TOpParams(Key1^).Types[1] > TOpParams(Key2^).Types[1] then Compare:=+1 else
         Compare:=0;
      end;
   end;
end;

{ TConvsCollection }

function  TConvsCollection.Compare;
begin
   if TConvParams(Key1^).TypeI < TConvParams(Key2^).TypeI then Compare:=-1 else
   if TConvParams(Key1^).TypeI > TConvParams(Key2^).TypeI then Compare:=+1 else
   if TConvParams(Key1^).TypeO < TConvParams(Key2^).TypeO then Compare:=-1 else
   if TConvParams(Key1^).TypeO > TConvParams(Key2^).TypeO then Compare:=+1 else
   Compare:=0;
end;

{ TExpressionCompiler }

constructor TExpressionCompiler.Init;
begin
   Inherited Init;

   Optimization:=True;
   Code:=@NoCode;

   New(Types, Init(4, 4));
   New(Ops, Init(4, 4)); Ops^.Duplicates:=True;
   New(Convs, Init(4, 4));
   New(Vars, Init(4, 4));

   AddType(tiNone, 0, nil);

   AddOp(diEndExpr, tiNone, tiNone, tiNone, -2, nil);
   AddOp('(', tiNone, tiNone, tiNone, -1, nil);
end;

destructor TExpressionCompiler.Done;
begin
   Inherited Done;

   Dispose(Types, Done);
   Dispose(Ops, Done);
   Dispose(Convs, Done);
   Dispose(Vars, Done);

   FreeCodeMem;
end;

procedure TExpressionCompiler.FreeCodeMem;
begin
   if (Code<>@NoCode) and (Code<>nil) then FreeMem(Code, CodeSize);
   Code:=@NoCode;
end;

function  TExpressionCompiler.FindType;
var
   Index           :Integer;
begin
   FindType:=-1;
   for Index:=0 to Types^.Count-1 do
   if TTypeParams(Types^.At(Index)^).Id = Id then
   begin
      FindType:=Index; Exit;
   end;
end;

procedure TExpressionCompiler.AddType;
var
   AType           :^TTypeParams;
begin
   if FindType(AnId)>=0 then Status:=ecDuplicate else
   begin
      if ASize>12 then Status:=ecError else
      begin
         New(AType, Init(AnId, ASize, AGetValue));
         Types^.Insert(AType);
      end;
   end;
end;

procedure TExpressionCompiler.AddVar;
var
   AVarP           :^TVarParams;
   Index           :Integer;
begin
   GCaseSensitive:=CaseSensitive;
   New(AVarP, Init(AName, AVar, FindType(AType)));
   if AVarP^.TypeV>=0 then
   begin
      if Vars^.Search(AVarP, Index) then
      begin
         Dispose(AVarP, Done); Status:=ecDuplicate;
      end else Vars^.Insert(AVarP);
   end else
   begin
      Dispose(AVarP, Done); Status:=ecUnknownId;
   end;
end;

procedure TExpressionCompiler.AddOp;
var
   AOp             :^TOpParams;
   Index, p1, p2   :Integer;
begin
   GCaseSensitive:=CaseSensitive;
   p1:=0;
   for p2:=1 to Length(AName)+1 do
   if (p2>Length(AName)) or (AName[p2]=' ') then
   begin
      New(AOp, Init(Copy(AName, p1+1, p2-p1-1),
         FindType(AType1), FindType(AType2), FindType(ATypeR),
         APrecedence, ACode));
      p1:=p2;

      if (AOp^.Types[0]>=0) and (AOp^.Types[1]>=0) and (AOp^.TypeR>=0) then
      begin
         Ops^.CmpMethod:=cmTypes;
         if Ops^.Search(AOp, Index) then
         begin
            Dispose(AOp, Done); Status:=ecDuplicate;
         end else Ops^.Insert(AOp);
         Ops^.CmpMethod:=cmName;
      end else
      begin
         Dispose(AOp, Done); Status:=ecUnknownId;
      end;
   end;
end;

procedure TExpressionCompiler.AddConvert;
var
   AConv           :^TConvParams;
begin
   New(AConv, Init(FindType(TypeI), FindType(TypeO), ACode));
   Convs^.Insert(AConv);
end;

var
   OpFunc          :Pointer;

procedure TExpressionCompiler.Compile;
var
   TypeRI          :Integer;

   i, dl, p, p1    :Integer;
   c, lc           :Char;
   PExpr,
   Stack, Id       :String;
   IdParams        :TIdParams;

   RealCodeSize,
   MaxCodeSize     :Word;
   RealCode        :Pointer;

procedure PushC(var Stack              :String;
                c                      :Char);
begin
   if Stack[0]=#255 then Status:=ecTooComplex else
   begin
      Inc(Stack[0]); Stack[Length(Stack)]:=c;
   end;
end;

function  PopC (var Stack              :String)  :Char;
begin
   if Stack='' then Status:=ecError else
   begin
      PopC:=Stack[Length(Stack)]; Dec(Stack[0]);
   end;
end;

procedure PushI(var Stack              :String;
                Id                     :Char;
                i                      :Word);
assembler;
asm
   les  di,Stack
   xor  bx,bx
   mov  bl,byte ptr es:[di]
   add  byte ptr es:[di],4
   jc   @@Error
   mov  al,Id
   mov  byte ptr es:[di+bx+1],al
   mov  byte ptr es:[di+bx+4],al
   mov  ax,i
   mov  word ptr es:[di+bx+2],ax
   jmp  @@Exit
@@Error:
   mov  byte ptr es:[di],0FFh
   les  di,Self
   mov  es:[di].TExpressionCompiler.Status,ecTooComplex
@@Exit:
end;

function  PopI (var Stack              :String)  :Word;
assembler;
asm
   les  di,Stack
   sub  byte ptr es:[di],2
   jc   @@Error
   xor  bx,bx
   mov  bl,byte ptr es:[di]
   mov  ax,word ptr es:[di+bx+1]
   jmp  @@Exit
@@Error:
   les  di,Self
   mov  es:[di].TExpressionCompiler.Status,ecError
   xor  ax,ax
@@Exit:
end;

procedure Pass1;
var
   LOp             :Boolean;

procedure ProcessId;
var
   Found           :Boolean;
   COp, SOp        :Integer;
   CVar            :Integer;

procedure ProcessOp;
var
   OpParams        :TOpParams;
begin
   if Id<>diEndExpr then
   begin
      OpParams.Name:=@Id; OpParams.Binary:=LOp;
      Ops^.CmpMethod:=cmKind;
      if not Ops^.Search(@OpParams, COp) then Status:=ecInvOpUsage;
      Ops^.CmpMethod:=cmName;
      if Status<>ecOk then Exit;
   end;

   repeat
      Found:=True;
      if Stack[Length(Stack)]=diOpIndex then
      begin
         PopC(Stack); SOp:=PopI(Stack); PopC(Stack);
         PushI(Stack, diOpIndex, SOp);
         if Status<>ecOk then Exit;
         if (TOpParams(Ops^.At(COp)^).Precedence < TOpParams(Ops^.At(SOp)^).Precedence) or
            ((TOpParams(Ops^.At(COp)^).Precedence = TOpParams(Ops^.At(SOp)^).Precedence) and
            ((TOpParams(Ops^.At(COp)^).Binary) or (Id=diEndExpr))) then
         begin
            PopC(Stack); PushI(PExpr, diOpIndex, PopI(Stack)); PopC(Stack);
            if Status<>ecOk then Exit;
            Found:=False;
         end;
      end;
   until Found;
   PushI(Stack, diOpIndex, COp);
   LOp:=(TOpParams(Ops^.At(COp)^).Types[0]=0);
end;

procedure ProcessVar;
begin
   PushI(PExpr, diVarIndex, CVar);
   LOp:=True;
end;

procedure ProcessValue;
begin
   if Id='' then Exit;
   PExpr:=PExpr + diValueASCII+Id[0]+Id+Id[0]+diValueASCII;
   LOp:=True;
end;

begin { ProcessId }
   if p1=0 then Exit;

   if p1<0 then Id:=diEndExpr else Id:=Copy(Expr, p1, i-p1-1);
   while (Id<>'') and (Id[1]=' ') do Delete(Id, 1, 1);
   while Id[Length(Id)]=' ' do Dec(Id[0]);
   IdParams.Name:=@Id;

   if Ops^.Search(@IdParams, COp) then ProcessOp else
   if Vars^.Search(@IdParams, CVar) then ProcessVar else
   ProcessValue;

   p1:=0;
end;

const
   IdChars         :Set of Char =
   ['0'..'9', '.', '$', 'A'..'Z', '_', 'a'..'z', #128..#255];

begin { Pass 1 }
   PushC(Expr, diEndExpr);
   PExpr:=''; Stack:=''; c:=#0; i:=1; p1:=0; LOp:=False;
   repeat
      lc:=c; c:=Expr[i]; Inc(i);

      if ((lc in IdChars) and not (c in IdChars)) or
         ((c in IdChars) and not (lc in IdChars)) or
         (c=' ') then ProcessId;
      case c of
         diEndExpr:
         begin
            ProcessId; p1:=-1; ProcessId;
         end;
         '(':
         begin
            ProcessId;
            PushC(Stack, c);
            LOp:=False;
         end;
         ')':
         begin
            if p1<>0 then
            begin
               Status:=ecError; Break;
            end;
            while Stack[Length(Stack)]<>'(' do
            begin
               if Stack[Length(Stack)]>=' ' then PushC(PExpr, PopC(Stack)) else
               begin
                  PExpr:=PExpr + Copy(Stack, Length(Stack)-(4-1), 4);
                  if Stack[0]>=#4 then Dec(Stack[0], 4) else Status:=ecError;
               end;
               if Stack='' then
               begin
                  Status:=ecError; Break;
               end;
            end;
            PopC(Stack);
         end;
         else
         begin
            if p1=0 then p1:=i-1;

            { Don't process '+' or '-' in a real value as an operation }
            case lc of
               '0'..'9', '.':
               if ((c='E') or (c='e')) and
                  ((Expr[i]='+') or (Expr[i]='-')) then Inc(i);
            end;
         end;
      end;
      if (Status<>ecOk) or (PExpr[0]=#255) then Break;
   until c=diEndExpr;
   if Status=ecOk then
   if PExpr[0]=#255 then Status:=ecTooComplex else
   if (Stack[0]<>#4) or (PExpr='') then Status:=ecError;
end;

procedure Pass2(DoWrite                :Boolean);
var
   Pos             :Integer;
   VarI            :Integer;
   VarP            :^TVarParams;
   ValueS          :String[4];
   ValueV          :record
      Size         :Byte;
      case Integer of
         0:(ValueW :Word);
         1:(ValueP :Pointer);
   end absolute ValueS;
   UseDS, UseSS    :Boolean;

   ParamsC         :Boolean;
   LCodePos        :Word;
   LStackPos       :Char;
   rAX, rBX, rCX,
   rDX, rSI, rDI   :Word;

   OpI, LOpI       :Integer;
   OpParams        :^TOpParams;

   SOpParams       :TOpParams;
   NeededTypes     :Array [-1..1] of Integer;
   OpFound         :Boolean;
   AdjustTypes     :Boolean;

   NewCodeSize     :Word;

procedure WriteCode(const S            :String);
begin
   NewCodeSize:=RealCodeSize + Length(S);
   if DoWrite then
   begin
      if NewCodeSize>MaxCodeSize
      then Status:=ecIntError
      else Move(S[1], (PChar(Code)+RealCodeSize)^, Length(S));
   end else
   if NewCodeSize>MaxCodeSize then MaxCodeSize:=NewCodeSize;
   RealCodeSize:=NewCodeSize;
end;

function  GetValue(const Str           :String;
                   var Val             :String;
                   TypeI               :Integer) :Boolean;
var
   GetValueProc    :TGetValueProc;
begin
   GetValueProc:=TTypeParams(Types^.At(TypeI)^).GetValue;

   if Assigned(GetValueProc)
   then GetValue:=GetValueProc(Str, Val)
   else GetValue:=True;
end;

procedure PushRegs(TypeI               :Integer);
begin
   case TTypeParams(Types^.At(TypeI)^).Size of
      1 ..2 : WriteCode(#$50);
      3 ..4 : WriteCode(#$52#$50);
      5 ..6 : WriteCode(#$52#$53#$50);
      7 ..8 : WriteCode(#$52#$51#$53#$50);
      9 ..10: WriteCode(#$56#$52#$51#$53#$50);
      11..12: WriteCode(#$57#$56#$52#$51#$53#$50);
   end;
end;

procedure PushValue(const Value        :String);
var
   Use286, Use386,
   UseShort        :Boolean;
   Value4          :LongInt;
begin
   Use286:=(Test8086<>0);
   Pos:=Length(Value);
   while Pos>0 do
   begin
      Use386:=(Test8086>=2) and (Pos>2);

      Value4:=PLongInt(PChar(@Value)+Pos-3)^;
      UseShort:=
         DoWrite and
         Use286 and
         ((Use386 and
         (((LongRec(Value4).Hi=0) and (LongRec(Value4).Lo<=$80)) or
         (LongRec(Value4).Hi=$FFFF) and (LongRec(Value4).Lo>=$FF80))) or
         (not Use386 and ((LongRec(Value4).Hi<=$80) or (LongRec(Value4).Hi>=$FF80))));

      if Use386 then WriteCode(#$66);
      if Use286 then
      if UseShort then WriteCode(#$6A) else WriteCode(#$68) else WriteCode(#$B8);

      if UseShort then
      if Use386 then WriteCode(Value[Pos-3]) else
      if Pos>1 then WriteCode(Value[Pos-1]) else WriteCode(Value[1])
      else
      begin
         if Use386 then
         begin
            WriteCode(Value[Pos-3]); WriteCode(Value[Pos-2]);
         end;
         if Pos>1 then WriteCode(Value[Pos-1]);
         WriteCode(Value[Pos]);
         if Pos<=1 then WriteCode(#0);
         if not Use286 then WriteCode(#$50);
      end;
      if Use386 then Dec(Pos, 2);
      Dec(Pos, 2);
   end;
end;

procedure InitVar(Index                :Integer);
begin
   VarP:=Vars^.At(Index);

   UseDS:=(LongRec(VarP^.Addr).Hi = DSeg);
   UseSS:=(LongRec(VarP^.Addr).Hi = SSeg);

   ValueV.Size:=2;
   if (not UseDS) and (not UseSS) then
   begin
      WriteCode(#$B8);
      ValueV.ValueW:=LongRec(VarP^.Addr).Hi;
      WriteCode(ValueS);
      WriteCode(#$8E#$C0);
   end;

   Pos:=TTypeParams(Types^.At(VarP^.TypeV)^).Size;
end;

procedure PushVar(Index                :Integer);
begin
   InitVar(Index);
   while Pos>0 do
   begin
      Dec(Pos, 2); ValueV.ValueW:=LongRec(VarP^.Addr).Lo+Pos;
      if UseSS then WriteCode(#$36) else
      if not UseDS then WriteCode(#$26);
      WriteCode(#$FF#$36);
      WriteCode(ValueS);
   end;
end;

procedure PushRegsC;
var
   Value           :String[2*6];
   Regs            :record
      Size         :Byte;
      Regs         :Array [0..5] of Word;
   end absolute Value;
begin
   Regs.Size:=TTypeParams(Types^.At(OpParams^.TypeR)^).Size;
   Regs.Regs[0]:=rAX;
   if Value[0]<=#4 then Regs.Regs[1]:=rDX else
   if Value[0]<=#6 then begin Regs.Regs[1]:=rBX; Regs.Regs[2]:=rDX; end else
   Move(rAX, Regs.Regs, SizeOf(Regs)-1);

   PushC(Stack, diValueInt);
   PushC(Stack, Value[0]);
   PushC(Stack, Char(OpParams^.TypeR));
   PushC(Stack, Char(Hi(OpParams^.TypeR)));
   Stack:=Stack + Value;
   PushC(Stack, Value[0]);
   PushC(Stack, diValueInt);
end;

procedure ValueToRegs(const Value      :String);
const
   RegCodes4       :Array [0..1] of Byte =
   ($B8, $BA);
   RegCodes6       :Array [0..2] of Byte =
   ($B8, $BB, $BA);
   RegCodes12      :Array [0..5] of Byte =
   ($B8, $BB, $B9, $BA, $BE, $BF);
var
   RegCodes        :PByteArray;
   RegIndex        :Integer;
begin
   Pos:=Length(Value);
   if Pos<=4 then RegCodes:=@RegCodes4 else
   if Pos<=6 then RegCodes:=@RegCodes6 else RegCodes:=@RegCodes12;
   RegIndex:=(Pos-1) shr 1;
   while Pos>0 do
   begin
      WriteCode(Char(RegCodes^[RegIndex])); Dec(RegIndex);
      if Pos>1 then WriteCode(Value[Pos-1]);
      WriteCode(Value[Pos]);
      if Pos<=1 then WriteCode(#0);
      Dec(Pos, 2);
   end;
end;

procedure VarToRegs(Index              :Integer);
const
   RegCodes4       :Array [0..1] of Byte =
   ($A1, $16);
   RegCodes6       :Array [0..2] of Byte =
   ($A1, $1E, $16);
   RegCodes12      :Array [0..5] of Byte =
   ($A1, $1E, $0E, $16, $36, $3E);
var
   RegCodes        :PByteArray;
   RegIndex        :Integer;
begin
   InitVar(Index);
   if Pos<=4 then RegCodes:=@RegCodes4 else
   if Pos<=6 then RegCodes:=@RegCodes6 else RegCodes:=@RegCodes12;
   RegIndex:=(Pos-1) shr 1;
   while Pos>0 do
   begin
      Dec(Pos, 2); ValueV.ValueW:=LongRec(VarP^.Addr).Lo+Pos;
      if UseSS then WriteCode(#$36) else
      if not UseDS then WriteCode(#$26);
      if RegIndex<>0 then WriteCode(#$8B);
      WriteCode(Char(RegCodes^[RegIndex])); Dec(RegIndex);
      WriteCode(ValueS);
   end;
end;

procedure Call(Addr                    :Pointer);
begin
   if Byte(Addr^)=$90 then
   begin
      Inc(PChar(Addr));
      while Byte(Addr^)<>$CB do
      begin
         WriteCode(Char(Addr^));
         Inc(PChar(Addr));
      end;
   end else
   begin
      WriteCode(#$9A);
      ValueV.Size:=4; ValueV.ValueP:=Addr;
      WriteCode(ValueS);
   end;
end;

procedure AdjustType(TypeI, TypeO      :Integer);
var
   ConvParams      :TConvParams;
   ConvI           :Integer;
begin
   ConvParams.TypeI:=TypeI;
   ConvParams.TypeO:=TypeO;
   if Convs^.Search(@ConvParams, ConvI)
   then Call(TConvParams(Convs^.At(ConvI)^).Code)
   else Status:=ecInvType;
end;

procedure ProcessOp;
var
   ParamI1, ParamI2,
   VarI            :Integer;
   Param1StackPos  :Char;
   AdjustSize      :Word;

   NewStack        :String absolute Expr;
   StackPosS,
   StackPosN       :Char;
   CodePosS        :Word;
   ParamI          :Integer;

procedure ProcessParam(ParamI          :Integer;
                       AdjustAllowed   :Boolean);
var
   TypeI           :Integer;
   ParamPos        :Integer;
   LCodePos        :Word;

procedure Adjust;
begin
   AdjustType(NeededTypes[ParamI], OpParams^.Types[ParamI]);
   PushRegs(OpParams^.Types[ParamI]);
end;

begin
   c:=PopC(Stack);
   case c of
      diValueASCII, diValueInt:
      begin
         lc:=PopC(Stack);
         if c=diValueASCII then
         begin
            if (ParamI<0) or
               GetValue(Copy(Stack, Length(Stack)-Byte(lc)+1, Byte(lc)), Id, OpParams^.Types[ParamI]) then
            begin
               NeededTypes[ParamI]:=MaxInt;
               for TypeI:=0 to Types^.Count-1 do
               if not GetValue(Copy(Stack, Length(Stack)-Byte(lc)+1, Byte(lc)), Id, TypeI) then
               begin
                  NeededTypes[ParamI]:=TypeI; Break;
               end;
               if NeededTypes[ParamI]=MaxInt then
               begin
                  Status:=ecInvValue;
                  NeededTypes[ParamI]:=OpParams^.Types[ParamI];
               end;
            end else NeededTypes[ParamI]:=OpParams^.Types[ParamI];
         end
         else Id:=Copy(Stack, Length(Stack)-Byte(lc)+1, Byte(lc));
         PushValue(Id);
         if Stack[0]>=lc then Dec(Stack[0], Byte(lc)) else Status:=ecError;
         if c=diValueInt then NeededTypes[ParamI]:=PopI(Stack);
         PopC(Stack);
      end;
      diVarIndex:
      begin
         VarI:=PopI(Stack);
         NeededTypes[ParamI]:=TVarParams(Vars^.At(VarI)^).TypeV;
         PushVar(VarI);
         ParamsC:=False;
      end;
      diStack:
      begin
         ParamPos:=PopI(Stack);
         PopC(Stack);
         NeededTypes[ParamI]:=PopI(Stack);
         ParamsC:=False;
      end;
      else
         Status:=ecIntError;
   end;
   PopC(Stack);

   AdjustSize:=0;
   if ParamI>=0 then
   if NeededTypes[ParamI] <> OpParams^.Types[ParamI] then
   if AdjustTypes then
   begin
      if AdjustAllowed then
      if c=diStack then
      begin
         LCodePos:=RealCodeSize;
         if DoWrite then
         begin
            DoWrite:=False; Adjust; DoWrite:=True;
            AdjustSize:=RealCodeSize-LCodePos;
            Move(
               (PChar(Code)+ParamPos)^,
               (PChar(Code)+ParamPos+AdjustSize)^,
               LCodePos-ParamPos);
            LCodePos:=RealCodeSize;
            RealCodeSize:=ParamPos;
            Adjust;
            RealCodeSize:=LCodePos;
         end else
         begin
            Adjust;
            AdjustSize:=RealCodeSize-LCodePos;
         end;
      end else Adjust
      else if c<>diStack then Adjust;
   end else Status:=ecInvType;
end;

begin
   ParamI1:=Ord(OpParams^.Types[0]=0);
   ParamI2:=Ord(OpParams^.Binary);

   ParamsC:=True; NeededTypes[0]:=0; NeededTypes[1]:=0;

   if ParamI1=ParamI2 then ProcessParam(ParamI1, True) else
   begin
      LCodePos:=RealCodeSize; LStackPos:=Stack[0];
      ProcessParam(ParamI2, True);
      RealCodeSize:=LCodePos+AdjustSize;
      ProcessParam(ParamI1, True);
      Param1StackPos:=Stack[0];
      Stack[0]:=LStackPos;
      ProcessParam(ParamI2, False);
      Stack[0]:=Param1StackPos;
   end;

   Call(OpParams^.Code);
(*
{
   NewStack:=''; ParamI:=0;
   StackPosS:=Stack[0]; CodePosS:=RealCodeSize;
   while (Stack<>'') and
         (Stack[Length(Stack[0])] in [diVarIndex, diValueASCII, diValueInt]) do
   begin
      Inc(ParamI);

      ProcessParam(-1, False);
      if ParamI>1+Ord(ParamI1<>ParamI2) then
      begin
         PushI(NewStack, diStack, NeededTypes[-1]);
         PopC(NewStack);
         PushI(NewStack, diStack, RealCodeSize);
      end else RealCodeSize:=CodePosS;
   end;
   StackPosN:=Stack[0]; Stack[0]:=StackPosS;
}
   ParamsC:=False{True};
   NeededTypes[0]:=0; NeededTypes[1]:=0;
{
   if ParamI1=ParamI2 then ProcessParam(ParamI1, True) else
   begin
      LCodePos:=RealCodeSize; LStackPos:=Stack[0];
      ProcessParam(ParamI2, True);
      RealCodeSize:=LCodePos+AdjustSize;
      ProcessParam(ParamI1, True);
      Param1StackPos:=Stack[0];
      Stack[0]:=LStackPos;
      ProcessParam(ParamI2, False);
      Stack[0]:=Param1StackPos;
   end;
}

   NewStack:='';
   while Stack<>'' do
   begin
      case Stack[1] of
         diVarIndex:
            Stack:=Copy(Stack, 5, 255)+Copy(Stack, 1, 4);
         diStack:
            {Stack:=Copy(Stack, Byte(Stack[2])+(2+2+2+1), 255)+Copy(Stack, 1, Byte(Stack[2])+(2+2+2));}
         begin
{            NewStack:=NewStack+Copy(Stack, 1, Byte(Stack[2])+(2+2+2));}
            Delete(Stack, 1, Byte(Stack[2])+(2+2+2));
            Continue;
         end
         else
            Stack:=Copy(Stack, Byte(Stack[2])+(2+2+1), 255)+Copy(Stack, 1, Byte(Stack[2])+(2+2));
      end;
      ProcessParam(-1, False);

      PushI(NewStack, diStack, NeededTypes[-1]);
      PopC(NewStack);
      PushI(NewStack, diStack, RealCodeSize);
   end;
   Stack:=NewStack;

   Call(OpParams^.Code);
{
   Stack[0]:=StackPosN; Stack:=Stack+NewStack;} *)
end;

procedure SelectOp;
var
   TryOpI          :Integer;
begin
   LCodePos:=RealCodeSize; LStackPos:=Stack[0];
   AdjustTypes:=False;
   repeat
      Status:=ecOk;
      ProcessOp;
      if Status=ecInvType then
      begin
         RealCodeSize:=LCodePos; Stack[0]:=LStackPos;
         SOpParams.Name:=OpParams^.Name;
         SOpParams.Binary:=OpParams^.Binary;
         SOpParams.Types[0]:=NeededTypes[0];
         SOpParams.Types[1]:=NeededTypes[1];
         Ops^.CmpMethod:=cmTypes;
         LOpI:=OpI;
         OpFound:=Ops^.Search(@SOpParams, OpI);
         Ops^.CmpMethod:=cmName;
         if OpFound then
         begin
            OpParams:=Ops^.At(OpI);
            if OpI=LOpI then Status:=ecIntError;
         end else
         if AdjustTypes then
         begin
            Ops^.CmpMethod:=cmKind;
            for TryOpI:=0 to Ops^.Count-1 do
            begin
               OpParams:=Ops^.At(TryOpI);
               if Ops^.Compare(OpParams, @SOpParams)=0 then
               begin
                  RealCodeSize:=LCodePos; Stack[0]:=LStackPos;
                  Status:=ecOk;
                  ProcessOp;
                  if Status<>ecInvType then Break;
               end;
            end;
            Ops^.CmpMethod:=cmName;
            Exit;
         end else AdjustTypes:=True;
      end;
   until Status<>ecInvType;
end;

begin { Pass 2 }
   if not DoWrite then MaxCodeSize:=0;
   RealCodeSize:=0;

   if SafeMode then WriteCode(#$55#$89#$E5);

   Stack:=''; i:=1;
   while i<=Length(PExpr) do
   begin
      c:=PExpr[i]; Inc(i);
      case c of
         diVarIndex, diValueASCII, diValueInt:
         begin
            if c<>diVarIndex then dl:=Byte(PExpr[i])+4 else dl:=4;
            Stack:=Stack + Copy(PExpr, i-1, dl);
            Inc(i, dl-1);
         end;
         diOpIndex:
         begin
            OpI:=PInteger(PChar(@PExpr)+i)^; Inc(i, 3);
            OpParams:=Ops^.At(OpI);

            SelectOp;
            if Status<>ecOk then Exit;

            if ParamsC and Optimization then
            begin
               WriteCode(#$89#$EC#$CB);
               if DoWrite then
               begin
                  OpFunc:=PChar(Code)+LCodePos;
                  asm
                     push bp
                     mov  bp,sp
                     sub  bp,4
                     call dword ptr OpFunc
                     pop  bp
                     mov  rAX,ax
                     mov  rBX,bx
                     mov  rCX,cx
                     mov  rDX,dx
                     mov  rSI,si
                     mov  rDI,di
                  end;
               end;

               RealCodeSize:=LCodePos;

               PushRegsC;
               Continue;
            end else
            if i<=Length(PExpr) then
            begin
               PushRegs(OpParams^.TypeR);
               PushI(Stack, diStack, OpParams^.TypeR);
               PopC(Stack);
               PushI(Stack, diStack, RealCodeSize);
            end else
            if OpParams^.TypeR<>TypeRI then
            begin
               PushRegs(OpParams^.TypeR);
               AdjustType(OpParams^.TypeR, TypeRI);
            end;
         end;
         else
            Status:=ecError;
      end;
      if Status<>ecOk then Exit;
   end;

   if Stack<>'' then
   begin
      NeededTypes[0]:=TypeRI;

      c:=PopC(Stack);
      if ((c=diValueASCII) and (Length(Stack)-Byte(PopC(Stack))=2)) or
         ((c=diValueInt) and (Length(Stack)-Byte(PopC(Stack))=2+2)) then
      begin
         if c=diValueASCII then
         begin
            if GetValue(Copy(Stack, 3, 255), Id, TypeRI) then Status:=ecInvValue;
         end
         else
         begin
            Id:=Copy(Stack, 3+2, 255);
            Stack[0]:=#4;
            NeededTypes[0]:=PopI(Stack);
         end;
         ValueToRegs(Id);
      end else
      if (c=diVarIndex) and (Stack[0]=#3) then
      begin
         VarI:=PopI(Stack);
         VarToRegs(VarI);
         NeededTypes[0]:=VarP^.TypeV;
      end else Status:=ecError;

      if NeededTypes[0]<>TypeRI then
      begin
         PushRegs(NeededTypes[0]);
         AdjustType(NeededTypes[0], TypeRI);
      end;
   end;

   if SafeMode then WriteCode(#$89#$EC#$5D);
   WriteCode(#$CB);
end;

procedure DoCompile;
begin
   TypeRI:=FindType(TypeR);
   if TypeRI<0 then
   begin
      Status:=ecUnknownId; Exit;
   end;

   Pass1; if Status<>ecOk then Exit;
   Pass2(False); if Status<>ecOk then Exit;

   GetMem(Code, MaxCodeSize);
   Pass2(True);

   if RealCodeSize>MaxCodeSize then Status:=ecIntError else
   begin
      GetMem(RealCode, RealCodeSize);
      Move(Code^, RealCode^, RealCodeSize);
      FreeMem(Code, MaxCodeSize);
      Code:=RealCode; CodeSize:=RealCodeSize;
   end;
end;

begin { Compile }
   GCaseSensitive:=CaseSensitive;

   Status:=ecOk;
   FreeCodeMem;
   DoCompile;
   if Status<>ecOk then FreeCodeMem;
end;

{ Standard types definition }

{$S-,R-,Q-}

{ Integer }

function  GetValueI(const Str          :String;
                    var Val            :String)  :Boolean; far;
var
   Error           :Integer;
   ValRec          :record
      Size         :Byte;
      Val          :LongInt;
   end absolute Val;

procedure GetHex(const Str             :String);
begin
   System.Val('$'+Str, ValRec.Val, Error);
end;

begin
   if (Str[1]='0') and (UpCase(Str[2])='X') then GetHex(Copy(Str, 3, 255)) else
   if UpCase(Str[Length(Str)])='H' then GetHex(Copy(Str, 1, Length(Str)-1)) else
   System.Val(Str, ValRec.Val, Error);
   ValRec.Size:=2;
   GetValueI:=(Error<>0) or (LongRec(ValRec.Val).Hi<>0);
end;

function  OpAddI   :Integer; far;
assembler;
asm
   nop
   pop  ax
   pop  dx
   add  ax,dx
end;

function  OpSubI   :Integer; far;
assembler;
asm
   nop
   pop  dx
   pop  ax
   sub  ax,dx
end;

function  OpMulI   :Integer; far;
assembler;
asm
   nop
   pop  ax
   pop  dx
   imul dx
end;

function  OpDivI   :Integer; far;
assembler;
asm
   nop
   pop  cx
   pop  ax
   cwd
   idiv cx
end;

function  OpModI   :Integer; far;
assembler;
asm
   nop
   pop  cx
   pop  ax
   cwd
   idiv cx
   mov  ax,dx
end;

function  OpCSgnI  :Integer; far;
assembler;
asm
   nop
   pop  ax
   neg  ax
end;

function  OpOrI    :Integer; far;
assembler;
asm
   nop
   pop  dx
   pop  ax
   or   ax,dx
end;

function  OpXorI   :Integer; far;
assembler;
asm
   nop
   pop  dx
   pop  ax
   xor  ax,dx
end;

function  OpAndI   :Integer; far;
assembler;
asm
   nop
   pop  dx
   pop  ax
   and  ax,dx
end;

function  OpShlI   :Integer; far;
assembler;
asm
   nop
   pop  cx
   pop  ax
   shl  ax,cl
end;

function  OpShrI   :Integer; far;
assembler;
asm
   nop
   pop  cx
   pop  ax
   shr  ax,cl
end;

function  OpNotI   :Integer; far;
assembler;
asm
   nop
   pop  ax
   not  ax
end;

function  OpI2I    :Integer; far;
assembler;
asm
   nop
   pop  ax
end;

procedure TExpressionCompiler.DefineInteger;
begin
   AddType(tiInteger, 2, GetValueI);

   AddOp('+', tiInteger, tiInteger, tiInteger, 10, @OpAddI);
   AddOp('-', tiInteger, tiInteger, tiInteger, 10, @OpSubI);
   AddOp('*', tiInteger, tiInteger, tiInteger, 20, @OpMulI);
   AddOp('div', tiInteger, tiInteger, tiInteger, 20, @OpDivI);
   AddOp('mod', tiInteger, tiInteger, tiInteger, 20, @OpModI);

   AddOp('-', tiNone, tiInteger, tiInteger, 30, @OpCSgnI);

   AddOp('or |', tiInteger, tiInteger, tiInteger, 10, @OpOrI);
   AddOp('xor', tiInteger, tiInteger, tiInteger, 10, @OpXorI);

   AddOp('and &', tiInteger, tiInteger, tiInteger, 20, @OpAndI);
   AddOp('shl <<', tiInteger, tiInteger, tiInteger, 20, @OpShlI);
   AddOp('shr >>', tiInteger, tiInteger, tiInteger, 20, @OpShrI);

   AddOp('not !', tiNone, tiInteger, tiInteger, 30, @OpNotI);

   AddOp('Integer', tiNone, tiInteger, tiInteger, 99, @OpI2I);
end;

{ Real }

function  GetValueR(const Str          :String;
                    var Val            :String)  :Boolean; far;
var
   Error           :Integer;
   ValRec          :record
      Size         :Byte;
      Val          :Real;
   end absolute Val;
begin
   System.Val(Str, ValRec.Val, Error);
   ValRec.Size:=SizeOf(Real);
   GetValueR:=(Error<>0);
end;

function  OpAddR(x, y                  :Real)    :Real; far;
begin
   OpAddR:=x+y;
end;

function  OpSubR(x, y                  :Real)    :Real; far;
begin
   OpSubR:=x-y;
end;

function  OpMulR(x, y                  :Real)    :Real; far;
begin
   OpMulR:=x*y;
end;

function  OpDivR(x, y                  :Real)    :Real; far;
begin
   OpDivR:=x/y;
end;

function  OpCSgnR(x                    :Real)    :Real; far;
begin
   OpCSgnR:=-x;
end;

function  OpPowerR(x, y                :Real)    :Real; far;
begin
   if y=0 then OpPowerR:=1 else
   if x=0 then OpPowerR:=0 else OpPowerR:=exp(ln(x)*y);
end;

function  OpSqrR(x                     :Real)    :Real; far;
begin
   OpSqrR:=sqr(x);
end;

function  OpSqrtR(x                    :Real)    :Real; far;
begin
   OpSqrtR:=sqrt(x);
end;

function  OpSinR(x                     :Real)    :Real; far;
begin
   OpSinR:=sin(x);
end;

function  OpCosR(x                     :Real)    :Real; far;
begin
   OpCosR:=cos(x);
end;

function  OpExpR(x                     :Real)    :Real; far;
begin
   OpExpR:=exp(x);
end;

function  OpLnR(x                      :Real)    :Real; far;
begin
   OpLnR:=ln(x);
end;

function  OpAbsR(x                     :Real)    :Real; far;
begin
   OpAbsR:=abs(x);
end;

function  OpArctanR(x                  :Real)    :Real; far;
begin
   OpArctanR:=arctan(x);
end;

function  OpTanR(x                     :Real)    :Real; far;
begin
   OpTanR:=sin(x)/cos(x);
end;

function  OpCTanR(x                    :Real)    :Real; far;
begin
   OpCTanR:=cos(x)/sin(x);
end;

function  OpLgR(x                      :Real)    :Real; far;
const
   ln10= 1/2.3025850930;
begin
   OpLgR:=ln10*ln(x);
end;

function  OpArcsinR(x                  :Real)    :Real; far;
begin
   if (x>=-1) and (x<=1) then
   if abs(x)=1
   then OpArcsinR:=x*(pi*0.5)
   else OpArcsinR:=arctan(x/sqrt(1-sqr(x))) else RunError(207);
end;

function  OpArccosR(x                  :Real)    :Real; far;
const
   PiA   :Array[Boolean] of Real =
   (0, 0);
begin
   if (x>=-1) and (x<=1) then
   if x=0 then OpArccosR:=pi*0.5 else
   begin
      PiA[True]:=pi;
      OpArccosR:=arctan(sqrt(1-sqr(x))/x)+PiA[x<0];
   end else RunError(207);
end;

procedure TExpressionCompiler.DefineReal;
begin
   AddType(tiReal, SizeOf(Real), GetValueR);

   AddOp('+', tiReal, tiReal, tiReal, 10, @OpAddR);
   AddOp('-', tiReal, tiReal, tiReal, 10, @OpSubR);
   AddOp('*', tiReal, tiReal, tiReal, 20, @OpMulR);
   AddOp('/', tiReal, tiReal, tiReal, 20, @OpDivR);

   AddOp('-', tiNone, tiReal, tiReal, 30, @OpCSgnR);

   AddOp('^', tiReal, tiReal, tiReal, 35, @OpPowerR);

   AddOp('sqr', tiNone, tiReal, tiReal, 40, @OpSqrR);
   AddOp('sqrt', tiNone, tiReal, tiReal, 40, @OpSqrtR);
   AddOp('sin', tiNone, tiReal, tiReal, 40, @OpSinR);
   AddOp('cos', tiNone, tiReal, tiReal, 40, @OpCosR);
   AddOp('exp', tiNone, tiReal, tiReal, 40, @OpExpR);
   AddOp('ln log', tiNone, tiReal, tiReal, 40, @OpLnR);
   AddOp('abs', tiNone, tiReal, tiReal, 40, @OpAbsR);
   AddOp('arctan atan', tiNone, tiReal, tiReal, 40, @OpArctanR);
   AddOp('tan tg', tiNone, tiReal, tiReal, 40, @OpTanR);
   AddOp('cotan ctg', tiNone, tiReal, tiReal, 40, @OpCTanR);
   AddOp('lg log10', tiNone, tiReal, tiReal, 40, @OpLgR);
   AddOp('arcsin asin', tiNone, tiReal, tiReal, 40, @OpArcsinR);
   AddOp('arccos acos', tiNone, tiReal, tiReal, 40, @OpArccosR);
end;

{ Integer <-> Real }

function  OpI2R(x                      :Integer) :Real; far;
begin
   OpI2R:=x;
end;

function  OpR2I(x                      :Real)    :Integer; far;
begin
   OpR2I:=Round(x);
end;

function  OpTrunc(x                    :Real)    :Integer; far;
begin
   OpTrunc:=Trunc(x);
end;

procedure TExpressionCompiler.DefineIR;
begin
   DefineInteger;
   DefineReal;

   AddOp('Real', tiNone, tiInteger, tiReal, 99, @OpI2R);
   AddOp('Integer', tiNone, tiReal, tiInteger, 99, @OpR2I);
   AddOp('round', tiNone, tiReal, tiInteger, 40, @OpR2I);
   AddOp('trunc', tiNone, tiReal, tiInteger, 40, @OpTrunc);

   AddConvert(tiInteger, tiReal, @OpI2R);
end;

{ Boolean }

function  GetValueB(const Str          :String;
                    var Val            :String)  :Boolean; far;
var
   ValRec          :record
      Size         :Byte;
      Val          :Byte;
   end absolute Val;
begin
   GetValueB:=False;
   Val:=Str; SLoCase(Val);
   if Val='true' then ValRec.Val:=$FF else
   if Val='false' then ValRec.Val:=0 else GetValueB:=True;
   ValRec.Size:=1;
end;

function  OpEqB    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dl,cl
   jne  @@1
   not  ax
@@1:
end;

function  OpNEqB   :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dl,cl
   je   @@1
   not  ax
@@1:
end;

function  OpLTB    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dl,cl
   jle  @@1
   not  ax
@@1:
end;

function  OpGTB    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dl,cl
   jge  @@1
   not  ax
@@1:
end;

function  OpLEB    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dl,cl
   jl   @@1
   not  ax
@@1:
end;

function  OpGEB    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dl,cl
   jg   @@1
   not  ax
@@1:
end;

procedure TExpressionCompiler.DefineBoolean;
begin
   AddType(tiBoolean, 1, GetValueB);

   AddOp('or || |', tiBoolean, tiBoolean, tiBoolean, 10, @OpOrI);
   AddOp('xor', tiBoolean, tiBoolean, tiBoolean, 10, @OpXorI);

   AddOp('and && &', tiBoolean, tiBoolean, tiBoolean, 20, @OpAndI);

   AddOp('not !', tiNone, tiBoolean, tiBoolean, 30, @OpNotI);

   AddOp('= ==', tiBoolean, tiBoolean, tiBoolean, 8, @OpEqB);
   AddOp('<> !=', tiBoolean, tiBoolean, tiBoolean, 8, @OpNEqB);

   AddOp('<', tiBoolean, tiBoolean, tiBoolean, 9, @OpLTB);
   AddOp('>', tiBoolean, tiBoolean, tiBoolean, 9, @OpGTB);
   AddOp('<= =<', tiBoolean, tiBoolean, tiBoolean, 9, @OpLEB);
   AddOp('>= =>', tiBoolean, tiBoolean, tiBoolean, 9, @OpGEB);
end;

{ Integer <-> Boolean }

function  OpEqI    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dx,cx
   jne  @@1
   not  ax
@@1:
end;

function  OpNEqI   :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dx,cx
   je   @@1
   not  ax
@@1:
end;

function  OpLTI    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dx,cx
   jge  @@1
   not  ax
@@1:
end;

function  OpGTI    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dx,cx
   jle  @@1
   not  ax
@@1:
end;

function  OpLEI    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dx,cx
   jg   @@1
   not  ax
@@1:
end;

function  OpGEI    :Boolean; far;
assembler;
asm
   nop
   pop  cx
   pop  dx
   xor  ax,ax
   cmp  dx,cx
   jl   @@1
   not  ax
@@1:
end;

function  OpB2I    :Boolean; far;
assembler;
asm
   nop
   pop  ax
   xor  ah,ah
   or   al,al
   jz   @@1
   mov  al,1
@@1:
end;

function  OpI2B    :Boolean; far;
assembler;
asm
   nop
   pop  ax
   or   ax,ax
   jz   @@1
   mov  ax,0FFh
@@1:
end;

procedure TExpressionCompiler.DefineBI;
begin
   AddOp('= ==', tiInteger, tiInteger, tiBoolean, 8, @OpEqI);
   AddOp('<> !=', tiInteger, tiInteger, tiBoolean, 8, @OpNEqI);

   AddOp('<', tiInteger, tiInteger, tiBoolean, 9, @OpLTI);
   AddOp('>', tiInteger, tiInteger, tiBoolean, 9, @OpGTI);
   AddOp('<= =<', tiInteger, tiInteger, tiBoolean, 9, @OpLEI);
   AddOp('>= =>', tiInteger, tiInteger, tiBoolean, 9, @OpGEI);

   AddOp('Integer', tiNone, tiBoolean, tiInteger, 99, @OpB2I);
   AddOp('Boolean', tiNone, tiInteger, tiBoolean, 99, @OpI2B);
end;

{ Real -> Boolean }

function  OpEqR(x, y                   :Real)    :Boolean; far;
begin
   OpEqR:=x=y;
end;

function  OpNEqR(x, y                  :Real)    :Boolean; far;
begin
   OpNEqR:=x<>y;
end;

function  OpLTR(x, y                   :Real)    :Boolean; far;
begin
   OpLTR:=x<y;
end;

function  OpGTR(x, y                   :Real)    :Boolean; far;
begin
   OpGTR:=x>y;
end;

function  OpLER(x, y                   :Real)    :Boolean; far;
begin
   OpLER:=x<=y;
end;

function  OpGER(x, y                   :Real)    :Boolean; far;
begin
   OpGER:=x>=y;
end;

procedure TExpressionCompiler.DefineBR;
begin
   AddOp('= ==', tiReal, tiReal, tiBoolean, 8, @OpEqR);
   AddOp('<> !=', tiReal, tiReal, tiBoolean, 8, @OpNEqR);

   AddOp('<', tiReal, tiReal, tiBoolean, 9, @OpLTR);
   AddOp('>', tiReal, tiReal, tiBoolean, 9, @OpGTR);
   AddOp('<= =<', tiReal, tiReal, tiBoolean, 9, @OpLER);
   AddOp('>= =>', tiReal, tiReal, tiBoolean, 9, @OpGER);
end;

procedure TExpressionCompiler.DefineAll;
begin
   DefineIR;
   DefineBoolean;
   DefineBI;
   DefineBR;
end;

end.


